module Parser where

import Syntax

import Context

import DeltaContext

import IO

import Monad

-- only for hugs
--isEOF = hugsIsEOF

type Env = Maybe (DeltaContext, Document)

parse :: IO Document

parse = do
	   [doc] <- ps [[]] [] [emptyContext] [Nothing]
	   return doc


ps :: 
	[Document]	-> -- stack of parent dccuments
	--[Document] 			-> -- stack of parent documents
	[Tag] 				-> -- stack of parent tags
	[Context] 			-> -- stack of contexts
	[Env] 				-> -- stack of extra information
	IO [Document]
	
ps docs tags ct env = do 
                     	eof <- isEOF
		     	if eof then return (removeEnv env ct docs)
		        	else do 
			               c         <- getChar
				       case c of
				            '<' -> (do {(a,b,d,e) <- eatTag docs tags ct env; ps a b d e})
			   	            _   -> if not (getTT (head ct)) then ps (eatChar c (removeEnv env ct docs)) tags ct (setHeadNothing env)
					                              else (do
								               (e,eof') <- eatText [c]
									       if eof' then return (insert e (removeEnv env ct docs))
									              else (do
										                (d',t',c',e')<- eatTag (insert e (removeEnv env ct docs)) tags ct (setHeadNothing env)
										      		ps d' t' c' e'))
removeEnv :: [Env] -> [Context] -> [Document] -> [Document]
removeEnv (Nothing:_) _ docs = docs
removeEnv ((Just dcd):_) (ct:_) (d:ds) = (enter dcd ct d) : ds

insert :: El -> [Document] -> [Document]
insert e (h:b) = (e : h) : b

setHeadNothing :: [Maybe a] -> [Maybe a]
setHeadNothing l@(Nothing:_) = l
setHeadNothing (_:l) = Nothing:l

eatText :: String -> IO (El,Bool)
eatText l = 
	do 
	   eof <- isEOF
	   if eof then return (Txt l, True)
	          else do
		         c <- getChar
			 case c of 
			      '<' -> return (Txt l, False)
			      _   -> eatText (c:l)



eatChar :: Char -> [Document] -> [Document] 

eatChar ' ' docs@((Ws:_):_) = docs -- compact whitespace
eatChar ' '  (h:t) = (Ws : h) : t 
eatChar '\n' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\n' (h:t) = (Ws : h) : t
eatChar '\t' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\t' (h:t) = (Ws : h) : t
eatChar '\r' docs@((Ws:_):_) = docs -- compact whitespace
eatChar '\r' (h:t) = (Ws : h) : t

eatChar c ( ( (Txt l) : hr) : tr) =  ((Txt (c:l)) : hr) : tr
eatChar c (h:t) = ((Txt [c]) : h) : t

eatTag :: [Document] -> [Tag] -> [Context] -> [Env] -> IO ([Document],[Tag],[Context],[Env])
eatTag docs tags ct env =
  do 
     c <- getChar
     restTag
     return (case c of
       		'/' -> popTag docs tags ct env
       		'E' -> pushEM docs tags ct (Nothing:env)
       		'T' -> pushTT docs tags ct (Nothing:env)
       		'P' -> pushPlain docs tags ct (Nothing:env)
       		_   -> pushTag c docs tags ct (Nothing:env))
       
popTag :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])

popTag ([]:ds) (_:ts) (_:cs) (Nothing:es) = -- do not generate teg for empty document 
	(ds,ts,cs,es)
	
popTag ([]:ds) (t:ts) (c1:c2:cs) (Just (dc,d):Nothing:es) =
	if c1 == c2 then (ds, ts, c2:cs, Just (dc,d):es)
	            else (ds, ts, c2:cs, Just (enterDCTag t dc,d):es)

popTag ([]:d:ds) (t:ts) (c1:c2:cs) (Just (dc1,d1) : Just (dc2,d2) : es) = 
	let dc1' 	= if c1 == c2 then dc1 else enterDCTag t dc1
	    result	= mergeByDeltaContext (dc2,d2) (dc1',d1) c2
	in case result of
		Left  (p2, p1) -> ((enter p2 c2 d): ds,ts,c2:cs,(Just p1):es)
		Right p3       -> (d:ds,ts,c2:cs,(Just p3):es)

popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Nothing:es) =
	if c1 == c2 then -- the current tqg does not add anything to the context
			 ((d1 ++ d2):ds, ts, c2:cs, Nothing:es)
	            else -- the current tqg does add something to the context
		         (d2:ds, ts, c2:cs,Just (enterDCTag t emptyDeltaContext,d1):es )

popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Just (dc,d) : Nothing : es) =
	if c1 == c2 then -- the current tag does not add anything to the context
			 ((d1 ++ d2):ds, ts, c2:cs, Just (dc,d) : es)
	 	    else -- the current tag does add something to the context
		         (d2:ds, ts, c2:cs, Just (enterDCTag t emptyDeltaContext, enter (dc,d) c1 d1) : es)

popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Nothing:Just (dc,d):es) =
	if c1 == c2 then
			 ((d1 ++ (enter (dc,d) c2 d2)):ds,ts,c2:cs, Nothing:es)
		    else
		    	 let result = mergeByDeltaContext (dc,d) (enterDCTag t emptyDeltaContext, d1) c2
			 in case result of
			         Left  (p2, p1) -> ((enter p2 c2 d2):ds,ts,c2:cs,Just p1:es)
				 Right p3       -> (d2:ds,ts,c2:cs,Just p3 : es)

popTag (d1:d2:ds) (t:ts) (c1:c2:cs) (Just (dc1,doc1) : Just (dc2,doc2) : es) =
	if c1 == c2 then
			 ((d1 ++ (enter (dc2,doc2) c2 d2)):ds , ts, c2:cs,Just (dc1, doc1) : es)
	            else
		         let result = mergeByDeltaContext (dc2,doc2) (enterDCTag t emptyDeltaContext, enter (dc1, doc1) c1 d1) c2
			 in case result of
			 	 Left  (p2, p1) -> ((enter p2 c2 d2) : ds, ts, c2:cs, Just p1: es)
				 Right p3       -> (d2 : ds, ts, c2:cs, Just p3 : es)

{-
Nog te doen:
  De huidige neighbour bijhouden en daarmee proberen te mergen.
  Geen nieuwe DeltaContext -> Geen neighbour -> geen nieuwe neighbour, gewoon enteren
                           -> wel            -> oude neighbour enteren, gewoon enteren, geen nieuwe neighbour
  Wel nieuwe deltacontext (evt subneighbour enteren)
  	-> geen neighbour -> zelf wordt nieuwe neighbour
	-> wel neighbour -> proberen te enteren
				-> niet gelukt, neighbour enteren en zelf neighbour worden
				-> wel gelukt, merge wordt nieuwe neighbour
-}

enter :: (DeltaContext, Document) -> Context -> Document -> Document
enter (dc, doc) ct parent = let child = buildDCEl dc ct doc
                            in child ++ parent

pushEM :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushEM docs tags (c:cs) env = 
	([]:docs,EM:tags,(enterTag c EM):c:cs,env)
 
pushTT :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushTT docs tags (c:cs) env = 
	([]:docs,TT:tags,(enterTag c TT):c:cs,env)

pushPlain :: [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushPlain docs tags (c:cs) env = 
	([]:docs,PL:tags,(enterTag c PL):c:cs,env)
 
pushTag :: Char -> [Document] -> [Tag] -> [Context] -> [Env] -> ([Document],[Tag],[Context],[Env])
pushTag x docs tags (c:cs) env = 
	let t = getTag x
	in ([]:docs,t:tags,(enterTag c t):c:cs,env)

restTag :: IO ()
restTag = 
  do 
     c <- getChar
     case c of 
       '>' -> return ()
       _   -> restTag
